package net.knarcraft.blacksmith.command;

import net.knarcraft.blacksmith.config.blacksmith.GlobalBlacksmithSetting;
import net.knarcraft.blacksmith.config.blacksmith.BlacksmithNPCSetting;
import net.knarcraft.blacksmith.config.SettingValueType;
import net.knarcraft.blacksmith.util.InputParsingHelper;
import org.bukkit.command.Command;
import org.bukkit.command.CommandSender;
import org.bukkit.command.TabCompleter;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.ArrayList;
import java.util.List;

import static net.knarcraft.blacksmith.util.TabCompleteValuesHelper.getTabCompletions;
import static net.knarcraft.knarlib.util.TabCompletionHelper.filterMatchingContains;

/**
 * The tab completer for the command used for changing global configuration options
 */
public class BlackSmithConfigTabCompleter implements TabCompleter {

    @Nullable
    @Override
    public List<String> onTabComplete(@NotNull CommandSender sender, @NotNull Command command, @NotNull String label,
                                      @NotNull String[] args) {
        if (!sender.hasPermission("blacksmith.admin")) {
            return new ArrayList<>();
        }

        //Arguments: <setting> [new value/material or enchantment] []

        //Prevent tab-completion when typing messages with spaces
        if (skipCompletionForSpacedMessage(args) != null) {
            return new ArrayList<>();
        }

        if (args.length == 1) {
            List<String> availableCommands = new ArrayList<>();
            availableCommands.add("reload");
            for (BlacksmithNPCSetting setting : BlacksmithNPCSetting.values()) {
                availableCommands.add(setting.getCommandName());
            }
            for (GlobalBlacksmithSetting globalBlacksmithSetting : GlobalBlacksmithSetting.values()) {
                availableCommands.add(globalBlacksmithSetting.getCommandName());
            }
            return filterMatchingContains(availableCommands, args[0]);
        } else if (args.length == 2) {
            return tabCompleteCommandValues(args[0], args[1]);
        } else if (args.length == 3) {
            //Get per-material tab completions, or return nothing if an invalid setting was specified
            for (GlobalBlacksmithSetting globalBlacksmithSetting : GlobalBlacksmithSetting.values()) {
                if (globalBlacksmithSetting.getCommandName().equalsIgnoreCase(args[0])) {
                    return getPerTypeTabCompletions(globalBlacksmithSetting, args);
                }
            }
            return new ArrayList<>();
        }
        return null;
    }

    /**
     * Checks whether to tab-complete nothing because a message containing spaces is being written
     *
     * @param args <p>The arguments given by the user</p>
     * @return <p>Null if not writing a spaced message</p>
     */
    private List<String> skipCompletionForSpacedMessage(String[] args) {
        if (args.length > 2) {
            BlacksmithNPCSetting blacksmithNpcSetting = null;
            for (BlacksmithNPCSetting setting : BlacksmithNPCSetting.values()) {
                if (setting.getCommandName().equalsIgnoreCase(args[0])) {
                    blacksmithNpcSetting = setting;
                    break;
                }
            }
            if (blacksmithNpcSetting != null && blacksmithNpcSetting.getPath().startsWith("defaults.messages")) {
                return new ArrayList<>();
            }
        }
        return null;
    }

    /**
     * Gets tab-completions for a selected material or enchantment
     *
     * @param globalBlacksmithSetting <p>The global setting to get tab-completions for</p>
     * @param args          <p>The arguments given by the user</p>
     * @return <p>The tab-completions to show to the user</p>
     */
    private List<String> getPerTypeTabCompletions(GlobalBlacksmithSetting globalBlacksmithSetting, String[] args) {
        //Display possible tab-completions only if a valid enchantment or material is provided
        if (((globalBlacksmithSetting == GlobalBlacksmithSetting.BASE_PRICE ||
                globalBlacksmithSetting == GlobalBlacksmithSetting.PRICE_PER_DURABILITY_POINT) &&
                InputParsingHelper.matchMaterial(args[1]) != null) ||
                (globalBlacksmithSetting == GlobalBlacksmithSetting.ENCHANTMENT_COST &&
                        InputParsingHelper.matchEnchantment(args[1]) != null)) {
            return filterMatchingContains(getTabCompletions(globalBlacksmithSetting.getValueType()), args[2]);
        } else {
            return new ArrayList<>();
        }
    }

    /**
     * Tab completes the values available for the given command
     *
     * @param commandName  <p>The name of the used command</p>
     * @param commandValue <p>The command value used to filter tab-completions</p>
     * @return <p>Some valid options for the command's argument</p>
     */
    private List<String> tabCompleteCommandValues(String commandName, String commandValue) {
        if (commandName.equalsIgnoreCase("reload")) {
            return new ArrayList<>();
        }
        for (GlobalBlacksmithSetting globalBlacksmithSetting : GlobalBlacksmithSetting.values()) {
            if (globalBlacksmithSetting.getCommandName().equalsIgnoreCase(commandName)) {
                return getCompletions(globalBlacksmithSetting, commandValue);
            }
        }
        for (BlacksmithNPCSetting blacksmithNpcSetting : BlacksmithNPCSetting.values()) {
            if (blacksmithNpcSetting.getCommandName().equalsIgnoreCase(commandName)) {
                return filterMatchingContains(getTabCompletions(
                        blacksmithNpcSetting.getValueType()), commandValue);
            }
        }
        return null;
    }

    /**
     * Gets tab-completions for the given global setting and filters on the command value
     *
     * @param globalBlacksmithSetting <p>The global setting to get tab-completions for</p>
     * @param commandValue  <p>The command value used to filter between available tab-completions</p>
     * @return <p>The available tab-completions</p>
     */
    private List<String> getCompletions(GlobalBlacksmithSetting globalBlacksmithSetting, String commandValue) {
        List<String> returnValues = filterMatchingContains(getTabCompletions(globalBlacksmithSetting.getValueType()), commandValue);
        if (globalBlacksmithSetting == GlobalBlacksmithSetting.BASE_PRICE || globalBlacksmithSetting == GlobalBlacksmithSetting.PRICE_PER_DURABILITY_POINT) {
            returnValues.addAll(filterMatchingContains(getTabCompletions(SettingValueType.MATERIAL), commandValue));
        } else if (globalBlacksmithSetting == GlobalBlacksmithSetting.ENCHANTMENT_COST) {
            returnValues.addAll(filterMatchingContains(getTabCompletions(SettingValueType.ENCHANTMENT), commandValue));
        }
        return returnValues;
    }

}
