package net.knarcraft.blacksmith.config.blacksmith;

import net.citizensnpcs.api.util.DataKey;
import net.knarcraft.blacksmith.BlacksmithPlugin;
import net.knarcraft.blacksmith.config.SettingValueType;
import net.knarcraft.blacksmith.config.SmithPreset;
import net.knarcraft.blacksmith.util.ConfigHelper;
import net.knarcraft.blacksmith.util.InputParsingHelper;
import net.knarcraft.blacksmith.util.ItemHelper;
import org.bukkit.Material;
import org.bukkit.enchantments.Enchantment;
import org.bukkit.inventory.ItemStack;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;

/**
 * A class which keeps track of all Blacksmith settings/config values for one NPC
 */
public class BlacksmithNPCSettings {

    private final List<Material> reforgeAbleItems = new ArrayList<>();
    private final List<Enchantment> enchantmentBlocklist = new ArrayList<>();
    private final Map<BlacksmithNPCSetting, Object> currentValues = new HashMap<>();
    private final GlobalBlacksmithSettings globalBlacksmithSettings;

    /**
     * Instantiates a new "Settings" object
     */
    public BlacksmithNPCSettings(GlobalBlacksmithSettings globalBlacksmithSettings) {
        this.globalBlacksmithSettings = globalBlacksmithSettings;
    }

    /**
     * Loads variables from the given data key
     *
     * @param key <p>The data key to load variables from</p>
     */
    public void loadVariables(DataKey key) {
        for (BlacksmithNPCSetting setting : BlacksmithNPCSetting.values()) {
            if (key.keyExists(setting.getChildPath())) {
                currentValues.put(setting, key.getRaw(setting.getChildPath()));
            }
        }
        //Updates the list of reforge-able items/materials
        updateReforgeAbleItems();
        updateEnchantmentBlocklist();
    }

    /**
     * Saves variables to the given data key
     *
     * @param key <p>The data key to save variables to</p>
     */
    public void saveVariables(DataKey key) {
        for (BlacksmithNPCSetting setting : BlacksmithNPCSetting.values()) {
            key.setRaw(setting.getChildPath(), currentValues.get(setting));
        }
    }

    /**
     * Changes one setting to the given value
     *
     * @param setting  <p>The setting to change</p>
     * @param newValue <p>The new value of the setting</p>
     */
    public void changeSetting(BlacksmithNPCSetting setting, Object newValue) {
        if (setting.getValueType() == SettingValueType.STRING_LIST ||
                setting.getValueType() == SettingValueType.REFORGE_ABLE_ITEMS) {
            //Workaround to make sure it's treated as the correct type
            currentValues.put(setting, newValue == null ? null : ConfigHelper.asStringList(newValue));
        } else {
            currentValues.put(setting, newValue);
        }
        if (setting == BlacksmithNPCSetting.REFORGE_ABLE_ITEMS) {
            updateReforgeAbleItems();
        }
        if (setting == BlacksmithNPCSetting.ENCHANTMENT_BLOCKLIST) {
            updateEnchantmentBlocklist();
        }
    }

    /**
     * Gets the raw current value of a setting
     *
     * @param setting <p>The setting to get the value of</p>
     * @return <p>The current value of the setting</p>
     */
    public Object getRawValue(BlacksmithNPCSetting setting) {
        return currentValues.get(setting);
    }

    /**
     * Gets the message to display when the blacksmith is busy with another player
     *
     * @return <p>The busy with player message</p>
     */
    public String getBusyWithPlayerMessage() {
        return asString(BlacksmithNPCSetting.BUSY_WITH_PLAYER_MESSAGE);
    }

    /**
     * Gets the message to display when the blacksmith is busy with reforging an item
     *
     * @return <p>The busy reforging message</p>
     */
    public String getBusyReforgingMessage() {
        return asString(BlacksmithNPCSetting.BUSY_WITH_REFORGE_MESSAGE);
    }

    /**
     * Gets the message to use for displaying an item's cost
     *
     * @return <p>The message to use for displaying item cost</p>
     */
    public String getCostMessage() {
        return asString(BlacksmithNPCSetting.COST_MESSAGE);
    }

    /**
     * Gets the message to display when a blacksmith has been given an invalid item
     *
     * @return <p>The invalid item message</p>
     */
    public String getInvalidItemMessage() {
        return asString(BlacksmithNPCSetting.INVALID_ITEM_MESSAGE);
    }

    /**
     * Gets the message to display when the presented item is at full durability
     *
     * @return <p>The not damaged message</p>
     */
    public String getNotDamagedMessage() {
        return asString(BlacksmithNPCSetting.NOT_DAMAGED_MESSAGE);
    }

    /**
     * Gets the message to display when a blacksmith starts reforging an item
     *
     * @return <p>The start reforge message</p>
     */
    public String getStartReforgeMessage() {
        return asString(BlacksmithNPCSetting.START_REFORGE_MESSAGE);
    }

    /**
     * Gets the message to display when a blacksmith has successfully repaired an item
     *
     * @return <p>The reforge success message</p>
     */
    public String getSuccessMessage() {
        return asString(BlacksmithNPCSetting.SUCCESS_MESSAGE);
    }

    /**
     * Gets the message to display when a blacksmith has failed to repair an item
     *
     * @return <p>The reforge fail message</p>
     */
    public String getFailMessage() {
        return asString(BlacksmithNPCSetting.FAIL_MESSAGE);
    }

    /**
     * Gets the message to display when a player cannot afford re-forging an item
     *
     * @return <p>The insufficient funds message</p>
     */
    public String getInsufficientFundsMessage() {
        return asString(BlacksmithNPCSetting.INSUFFICIENT_FUNDS_MESSAGE);
    }

    /**
     * Gets the message to display when a blacksmith is still affected by a cool-down
     *
     * @return <p>The cool down unexpired message</p>
     */
    public String getCoolDownUnexpiredMessage() {
        return asString(BlacksmithNPCSetting.COOL_DOWN_UNEXPIRED_MESSAGE);
    }

    /**
     * Gets the message to display when a player has changed the item they're trying to reforge
     *
     * @return <p>The item changed message</p>
     */
    public String getItemChangedMessage() {
        return asString(BlacksmithNPCSetting.ITEM_UNEXPECTEDLY_CHANGED_MESSAGE);
    }

    /**
     * Gets all items reforge-able by this NPC
     *
     * <p>If this is not empty, only the items specified can be repaired by this NPC.</p>
     *
     * @return <p>All items reforge-able by this NPC</p>
     */
    public List<Material> getReforgeAbleItems() {
        Object currentValue = currentValues.get(BlacksmithNPCSetting.REFORGE_ABLE_ITEMS);
        if (currentValue == null || String.valueOf(currentValue).isEmpty()) {
            return globalBlacksmithSettings.getReforgeAbleItems();
        } else {
            return new ArrayList<>(this.reforgeAbleItems);
        }
    }

    /**
     * Gets the list of blocked enchantments
     *
     * @return <p>The list of blocked enchantments</p>
     */
    public List<Enchantment> getEnchantmentBlocklist() {
        Object currentValue = currentValues.get(BlacksmithNPCSetting.ENCHANTMENT_BLOCKLIST);
        if (currentValue == null || String.valueOf(currentValue).isEmpty()) {
            return globalBlacksmithSettings.getEnchantmentBlocklist();
        } else {
            return new ArrayList<>(this.enchantmentBlocklist);
        }
    }

    /**
     * Gets the minimum delay used to wait for a reforging to finish.
     *
     * @return <p>The minimum reforge delay</p>
     */
    public int getMinReforgeDelay() {
        return asInt(BlacksmithNPCSetting.MIN_REFORGE_DELAY);
    }

    /**
     * Gets the maximum delay used to wait for a reforging to finish
     *
     * @return <p>The maximum reforge delay</p>
     */
    public int getMaxReforgeDelay() {
        return asInt(BlacksmithNPCSetting.MAX_REFORGE_DELAY);
    }

    /**
     * Gets the cool-down between each reforge
     *
     * @return <p>The reforge cool-down</p>
     */
    public int getReforgeCoolDown() {
        return asInt(BlacksmithNPCSetting.REFORGE_COOL_DOWN);
    }

    /**
     * Gets the chance to fail a reforging
     *
     * @return <p>The fail chance</p>
     */
    public int getFailChance() {
        return asInt(BlacksmithNPCSetting.FAIL_CHANCE);
    }

    /**
     * Gets whether a failed reforging should remove/downgrade enchantments
     *
     * @return <p>Whether enchantments should be removed</p>
     */
    public boolean getFailRemovesEnchantments() {
        return asBoolean(BlacksmithNPCSetting.FAIL_REMOVE_ENCHANTMENTS);
    }

    /**
     * Gets the chance for adding an extra enchantment to an item
     *
     * @return <p>The extra enchantment chance</p>
     */
    public int getExtraEnchantmentChance() {
        return asInt(BlacksmithNPCSetting.EXTRA_ENCHANTMENT_CHANCE);
    }

    /**
     * Gets the max number of enchantment to add to an item
     *
     * @return <p>The maximum enchantments</p>
     */
    public int getMaxEnchantments() {
        return asInt(BlacksmithNPCSetting.MAX_ENCHANTMENTS);
    }

    /**
     * Gets whether an item should be dropped on the ground instead of being given to the player
     *
     * @return <p>Whether to drop reforged items on the ground</p>
     */
    public boolean getDropItem() {
        return asBoolean(BlacksmithNPCSetting.DROP_ITEM);
    }

    /**
     * Gets the title of the blacksmith, describing which items it can reforge
     *
     * @return <p>The title of the blacksmith</p>
     */
    public String getBlacksmithTitle() {
        return asString(BlacksmithNPCSetting.BLACKSMITH_TITLE);
    }

    /**
     * Gets whether to disable the reforge-cool-down
     *
     * @return <p>Whether to disable the reforge-cool-down</p>
     */
    public boolean getDisableCoolDown() {
        return asInt(BlacksmithNPCSetting.REFORGE_COOL_DOWN) <= 0;
    }

    /**
     * Gets whether to disable the delay between starting reforging and the reforging finishing
     *
     * @return <p>Whether to disable the reforge delay</p>
     */
    public boolean getDisableDelay() {
        return asInt(BlacksmithNPCSetting.MAX_REFORGE_DELAY) <= 0;
    }

    /**
     * Gets whether this blacksmith is able to repair anvils
     *
     * @return <p>True if this blacksmith is able to repair anvils</p>
     */
    public boolean getRepairAnvils() {
        return asBoolean(BlacksmithNPCSetting.REPAIR_ANVILS);
    }

    /**
     * Gets the given value as an integer
     *
     * <p>This will throw an exception if used for a non-integer setting</p>
     *
     * @param setting <p>The setting to get the value of</p>
     * @return <p>The value of the given setting as an integer</p>
     */
    private int asInt(BlacksmithNPCSetting setting) {
        return ConfigHelper.asInt(getValue(setting));
    }

    /**
     * Gets the string value of the given setting
     *
     * @param setting <p>The setting to get the value of</p>
     * @return <p>The value of the given setting as a string</p>
     */
    private String asString(BlacksmithNPCSetting setting) {
        return getValue(setting).toString();
    }

    /**
     * Gets the boolean value of the given setting
     *
     * @param setting <p>The setting to get the value of</p>
     * @return <p>The value of the given setting as a boolean</p>
     */
    private boolean asBoolean(BlacksmithNPCSetting setting) {
        return ConfigHelper.asBoolean(getValue(setting));
    }

    /**
     * Gets the value of a setting, using the default if not set
     *
     * @param setting <p>The setting to get the value of</p>
     * @return <p>The current value</p>
     */
    private Object getValue(BlacksmithNPCSetting setting) {
        Object value = currentValues.get(setting);
        //If not set, use the default value from the config.yml file
        if (value == null) {
            Map<BlacksmithNPCSetting, Object> defaultNPCSettings = globalBlacksmithSettings.getDefaultNPCSettings();
            if (defaultNPCSettings.containsKey(setting)) {
                value = defaultNPCSettings.get(setting);
            }
        }
        //If not set in config.yml, use the default value from the enum
        if (value == null) {
            value = setting.getDefaultValue();
        }
        return value;
    }

    /**
     * Replaces placeholders in the given reforge-able value
     *
     * @param stringList <p>The value specified by a user</p>
     * @return <p>The value with placeholders replaced</p>
     */
    private static List<String> replaceReforgeAblePresets(List<String> stringList) {
        List<String> newStrings = new ArrayList<>();
        for (String item : stringList) {
            if (item == null) {
                continue;
            }
            String replaced = SmithPreset.replacePreset(item);
            if (!replaced.equals(item)) {
                newStrings.addAll(List.of(replaced.split(",")));
            } else {
                newStrings.add(item);
            }
        }
        return newStrings;
    }

    /**
     * Updates the list of blocked enchantments
     */
    private void updateEnchantmentBlocklist() {
        this.enchantmentBlocklist.clear();
        this.enchantmentBlocklist.addAll(getEnchantmentBlocklist(ConfigHelper.asStringList(getValue(
                BlacksmithNPCSetting.ENCHANTMENT_BLOCKLIST))));
    }

    /**
     * Gets the list of enchantments listed in the given string list
     *
     * @param enchantments <p>The enchantment names to block</p>
     * @return <p>The enchantments to be blocked</p>
     */
    public static List<Enchantment> getEnchantmentBlocklist(List<String> enchantments) {
        List<Enchantment> enchantmentBlocklist = new ArrayList<>();

        for (String item : enchantments) {
            if (InputParsingHelper.isEmpty(item)) {
                continue;
            }

            Enchantment enchantment = InputParsingHelper.matchEnchantment(item);
            if (enchantment != null) {
                enchantmentBlocklist.add(enchantment);
            } else {
                BlacksmithPlugin.getInstance().getLogger().log(Level.WARNING, "Unable to verify " + item +
                        " as a valid enchantment");
            }
        }
        return enchantmentBlocklist;
    }

    /**
     * Updates the reforge-able items according to the current value of the setting
     */
    private void updateReforgeAbleItems() {
        this.reforgeAbleItems.clear();
        this.reforgeAbleItems.addAll(getReforgeAbleItems(ConfigHelper.asStringList(getValue(
                BlacksmithNPCSetting.REFORGE_ABLE_ITEMS))));
    }

    /**
     * Gets a list of the reforgeAbleItems described in the given item list
     *
     * @param itemList <p>The list of items defined by the user</p>
     * @return <p>The materials contained in the item list</p>
     */
    public static List<Material> getReforgeAbleItems(List<String> itemList) {
        List<Material> reforgeAbleItems = new ArrayList<>();
        if (itemList == null) {
            return null;
        }

        //Convert any presets with a list of materials
        itemList = replaceReforgeAblePresets(itemList);

        Set<Material> blacklisted = new HashSet<>();
        //Parse every material, and add to reforgeAble items
        for (String item : itemList) {
            //Ignore ,,
            if (InputParsingHelper.isEmpty(item)) {
                continue;
            }

            boolean blacklist = false;
            if (item.startsWith("-")) {
                blacklist = true;
                item = item.substring(1);
            }

            Material material = InputParsingHelper.matchMaterial(item);
            if (material != null && ItemHelper.isRepairable(new ItemStack(material, 1))) {
                if (!blacklist) {
                    reforgeAbleItems.add(material);
                } else {
                    blacklisted.add(material);
                }
            } else {
                BlacksmithPlugin.getInstance().getLogger().log(Level.WARNING, "Unable to verify " + item +
                        " as a valid reforge-able item");
            }
        }

        //Remove any blacklisted materials at the end to make sure order of arguments won't matter
        reforgeAbleItems.removeAll(blacklisted);
        return reforgeAbleItems;
    }

}